from click import Option
from ..utils.helper_functions import _invoke_layout
from ..components import ChildContainer, LacyContainer, SlotContainer

from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Awaitable, Optional
from uuid import UUID
import asyncio

from dash import html
from dash.development.base_component import Component


@dataclass
class ExecNode:
    """Represents a node in the execution tree"""

    segment: str
    node_id: str
    parent_id: str
    layout: Callable[..., Awaitable[Component]] | Component
    variables: Dict[str, str] = field(default_factory=dict)
    slots: Dict[str, "ExecNode"] = field(default_factory=dict)
    child_node: Optional["ExecNode"] = "default"
    loading: Optional[Callable | Component] = None
    error: Optional[Callable | Component] = None
    is_lacy: bool = False

    async def execute(self, endpoint_results: Dict[UUID, Dict[any, any]]) -> Component:
        """
        Executes the node by rendering its layout with the provided variables,
        slots, and views.
        """
        data = endpoint_results.get(self.node_id)

        if self.is_lacy:
            loading_layout = await _invoke_layout(self.loading, **self.variables)
            return LacyContainer(loading_layout, str(self.node_id), self.variables)

        if isinstance(data, Exception):
            return await self.handle_error(data, self.variables)

        slots_content, views_content = await asyncio.gather(
            self._handle_slots(endpoint_results),
            self._handle_child(endpoint_results),
        )

        all_kwargs = {**self.variables, **slots_content, **views_content, "data": data}

        try:
            layout = await _invoke_layout(self.layout, **all_kwargs)
        except Exception as e:
            layout = await self.handle_error(e, self.variables)

        return layout

    async def handle_error(self, error: Exception, variables: Dict[str, Any]):
        if not self.error:
            return html.Div(str(error), className="banner")

        error_layout = await _invoke_layout(self.error, error, **variables)
        return error_layout

    async def _handle_slots(
        self, endpoint_results: Dict[UUID, Dict[Any, Any]]
    ) -> Dict[str, Component]:
        """Executes all slot nodes and gathers their rendered components."""
        if not self.slots:
            return {}

        executables = [slot.execute(endpoint_results) for slot in self.slots.values()]
        views = await asyncio.gather(*executables)
        results = {}

        for slot_name, slot_layout in zip(self.slots.keys(), views):
            clean_slot_name = slot_name.strip("()")
            results[clean_slot_name] = SlotContainer(
                slot_layout, self.node_id, slot_name
            )

        return results

    async def _handle_child(
        self, endpoint_results: Dict[UUID, Dict[Any, Any]]
    ) -> Dict[str, Component]:
        """Executes the current view node."""
        if self.child_node == "default":
            return {}

        layout = await self.child_node.execute(endpoint_results) if self.child_node else None

        return {
            "children": ChildContainer(
                layout, self.node_id, self.child_node.segment if self.child_node else None
            )
        }



from typing import Callable, get_type_hints, get_origin
from pydantic import BaseModel
import inspect


def classify_type(obj):
    """Classify what kind of type this is"""
    if not inspect.isclass(obj) and get_origin(obj) is None:
        return "not_a_type"

    # Handle typing generics
    if hasattr(obj, "__module__"):
        if obj.__module__ == "builtins":
            return "builtin"
        elif obj.__module__ == "typing":
            return "typing"

    # Check if it's a Pydantic model
    if inspect.isclass(obj) and issubclass(obj, BaseModel):
        return "pydantic"

    return "custom_class"


def should_skip(obj):
    """Skip everything except builtins and pydantic models"""
    return classify_type(obj) not in ("builtin", "typing", "pydantic")


def extract_function_inputs(func: Callable):
    """Recursively extracts function & class inputs"""
    valid_classes = ("builtin", "typing", "pydantic")
    inputs = []
    types = {}

    if not func:
        return inputs, types

    for param_name, param_type in get_type_hints(func).items():
        type_class = classify_type(param_type)

        if type_class not in valid_classes:
            continue

        if type_class == "pydantic":
            model_inputs = extract_pydantic_fields(param_type)
            inputs += model_inputs
            types[param_name] = param_type
            continue

        types[param_name] = param_type
        inputs.append(param_name)

    return inputs, types


def extract_pydantic_fields(model_class):
    """Extract fields from a Pydantic model"""
    fields = []
    for field_name, _ in model_class.model_fields.items():
        fields.append(field_name)

    return fields



from typing import Dict
from uuid import UUID

from .core.routing import PageNode


class RouteTreeValidationError(Exception):
    pass


def validate_static_route(node: PageNode):
    if not node.is_static:
        return

    if node.child_nodes:
        raise RouteTreeValidationError(
            f"""
            Route node: {node.module} is marked as static but contains child nodes.
            Either remove the static setting or the child nodes.
            """
        )

    if node.slots:
        raise RouteTreeValidationError(
            f"""
            Route node: {node.module} is marked as static but contains slots.
            Either remove the static setting or the slots.
            """
        )


def validate_default_child(node: PageNode):
    if not node.default_child:
        return

    if node.default_child not in node.child_nodes:
        raise RouteTreeValidationError(
            f"""
            Route node: {node.module} has no child node with name {node.default_child}.
            Default childs have to be present as children.
            """
        )

    if node.default_child and not node.child_nodes:
        raise RouteTreeValidationError(
            f"""
            Route node: {node.module} has no child node. 
            Setting default child {node.default_child} is not needed.
            """
        )


def validate_slots(node: PageNode, route_table: Dict[UUID, PageNode]):
    path_templates = []

    for _, slot_id in node.slots.items():
        slot_node = route_table.get(slot_id)
        path_templates.append(slot_node.path_template)

    # Pass if all pathtemplates are None
    if all([path_template is None for path_template in path_templates]):
        return

    # Pass if every slot has a pathtemplate
    if all([path_template is not None for path_template in path_templates]):
        return

    if (
        len(
            [
                path_template if path_template else None
                for path_template in path_templates
            ]
        )
        == 1
    ):
        return

    raise RouteTreeValidationError(
        f"""
        Slot Validation Error in node {node.module}.
        Either all, none or one slot can be dynamic.
        """
    )


def validate_child_nodes():
    pass


def validate_slot_pathtemplates():
    pass



from click import Option
from ..utils.helper_functions import _invoke_layout
from ..components import ChildContainer, LacyContainer, SlotContainer

from dataclasses import dataclass, field
from typing import Any, AnyStr, Callable, Dict, Optional
from uuid import UUID

from dash import html
from dash.development.base_component import Component


@dataclass
class ExecNode:
    """Represents a node in the execution tree"""

    segment: str
    node_id: str
    parent_id: str
    layout: Callable[..., Component] | Component
    variables: Dict[str, str] = field(default_factory=dict)
    slots: Dict[str, "ExecNode"] = field(default_factory=dict)
    child_node: Optional["ExecNode"] = "default"
    loading: Optional[Callable | Component] = None
    error: Optional[Callable | Component] = None
    is_lacy: bool = False

    def execute(self, endpoint_results: Dict[UUID, Dict[str, Any]]) -> Component:
        """
        Executes the node by rendering its layout with the provided variables,
        slots, and views.
        """
        data = endpoint_results.get(self.node_id)

        if self.is_lacy:
            loading_layout = _invoke_layout(self.loading, **self.variables)
            return LacyContainer(loading_layout, str(self.node_id), self.variables)

        if isinstance(data, Exception):
            return self.handle_error(data, self.variables)

        slots_content = self._handle_slots(endpoint_results)
        views_content = self._handle_child(endpoint_results)

        all_kwargs = {**self.variables, **slots_content, **views_content, "data": data}

        try:
            layout = _invoke_layout(self.layout, **all_kwargs)
        except Exception as e:
            layout = self.handle_error(e, self.variables)

        return layout

    def handle_error(self, error: Exception, variables: Dict[str, Any]):
        if not self.error:
            return html.Div(str(error), className="banner")

        error_layout = _invoke_layout(self.error, error, **variables)
        return error_layout

    def _handle_slots(
        self, endpoint_results: Dict[UUID, Dict]
    ) -> Dict[str, Component]:
        """Executes all slot nodes and gathers their rendered components."""
        if not self.slots:
            return {}

        results = {}
        for slot_name, slot_node in self.slots.items():
            slot_layout = slot_node.execute(endpoint_results)
            clean_slot_name = slot_name.strip("()")
            results[clean_slot_name] = SlotContainer(
                slot_layout, self.node_id, slot_name
            )

        return results

    def _handle_child(
        self, endpoint_results: Dict[UUID, Dict]
    ) -> Dict[str, Component]:
        """Executes the current view node."""
        if self.child_node == "default":
            return {}

        layout = self.child_node.execute(endpoint_results) if self.child_node else None

        return {
            "children": ChildContainer(
                layout, self.node_id, self.child_node.segment if self.child_node else None
            )
        }



from dataclasses import dataclass, field
from functools import partial
from typing import Any, Callable, Dict, List, Literal, Optional
from uuid import UUID

from ..utils.constants import DEFAULT_LAYOUT_TOKEN, REST_TOKEN
from .routing import PageNode, RouteTable
from pydantic import BaseModel


class LoadingState(BaseModel):
    state: Literal["lacy", "done", "hidden"]
    node_id: str
    updated: bool = False

    def update_state(self, state: Literal["lacy", "done", "hidden"]) -> None:
        self.state = state
        self.updated = True


@dataclass
class RoutingContext:
    """Encapsulates all routing state for a single request"""

    pathname: str
    query_params: Dict[str, Any]
    resolve_type: Literal["search", "url", "lacy"]
    path_vars: Dict[str, str] = field(default_factory=dict)
    endpoints: Dict[UUID, Callable] = field(default_factory=dict)
    segments: List[str] = field(default_factory=list)
    _loading_states: Dict[str, Any] = field(default_factory=dict, repr=False)

    @property
    def variables(self):
        return {**self.query_params, **self.path_vars}

    @classmethod
    def from_request(
        cls,
        pathname: str,
        query_params: Dict[str, Any],
        loading_state_dict: Dict[str, Dict],
        resolve_type: Literal["search", "url", "lacy"],
    ) -> "RoutingContext":
        """Create context from request data"""
        path = pathname.strip("/")
        segments = [seg for seg in path.split("/") if seg] if path else []
        # Build _loading_states dict
        _loading_states = {
            segment_key: LoadingState(**ils)
            for segment_key, ils in loading_state_dict.items()
        }
        return cls(
            pathname=pathname,
            query_params=query_params,
            segments=segments,
            resolve_type=resolve_type,
            _loading_states=_loading_states,
        )

    def get_node_state(self, segment_key: Optional[str] = None) -> Optional[str]:
        """Get loading state for a node"""
        ls = self._loading_states.get(segment_key)
        if ls:
            return ls.state
        return None

    def set_node_state(
        self, node: PageNode, state: str, segment_key: Optional[str] = None
    ):
        """Set loading state for a node"""
        if segment_key not in self._loading_states:
            self._loading_states[segment_key] = LoadingState(
                state=state, node_id=node.node_id, updated=True
            )
        else:
            self._loading_states[segment_key].update_state(state)

    def add_endpoint(self, node: PageNode):
        partial_endpoint = partial(node.endpoint, **self.variables)
        self.endpoints[node.node_id] = partial_endpoint

    def should_lazy_load(
        self, node: PageNode, segment_key: Optional[str] = None
    ) -> bool:
        """Check if node should be lazy loaded"""
        return (
            node.loading is not None
            and self.resolve_type != "lacy"
            and DEFAULT_LAYOUT_TOKEN not in segment_key
        )

    def pop_segment(self) -> Optional[str]:
        """Remove and return the last segment"""
        return self.segments.pop() if self.segments else None

    def peek_segment(self) -> Optional[str]:
        """Peek at the last segment without removing"""
        return self.segments[-1] if self.segments else None

    def consume_path_var(self, node: PageNode) -> Optional[str]:
        """Consume a segment as path variable"""
        if not node.is_path_template:
            return self.peek_segment()
        
        if node.segment == REST_TOKEN:
            rest_value = list(reversed(self.segments))
            self.segments = []
            self.path_vars["rest"] = rest_value
            return None
        else:
            value = self.pop_segment()
            if value:
                self.path_vars[node.segment] = value
            return value

    def merge_segments(self, ignore_empty_folders: bool):
        """Merge segments if empty folder should be ignored"""
        if ignore_empty_folders:
            self.pop_segment()
            return

        if len(self.segments) < 2:
            self.pop_segment()
            return

        first = self.segments.pop()
        second = self.segments.pop()
        combined = f"{first}/{second}"
        self.segments.append(combined)

    def gather_endpoints(self):
        if not self.endpoints:
            return {}

        keys = list(self.endpoints.keys())
        funcs = list(self.endpoints.values())
        results = []
        for func in funcs:
            try:
                result = func()
                results.append(result)
            except Exception as e:
                results.append(e)
        return dict(zip(keys, results))

    def to_loading_state_dict(self) -> Dict[str, Any]:
        """Convert context back to loading state dict for response"""
        return {**self.get_updated_loading_state(), "query_params": self.query_params}

    def get_updated_loading_state(self) -> Dict[str, Dict]:
        """Return only updated loading states as a dict."""
        return {
            key: {"state": state.state, "node_id": state.node_id}
            for key, state in self._loading_states.items()
            if state.updated == True
        }

    def set_silent_loading_states(self, node: PageNode, state: str = "done"):
        """Mark all descendant slots as done"""
        for slot_name, slot_id in node.slots.items():
            slot_node = RouteTable.get_node(slot_id)
            if slot_node:
                self.set_node_state(slot_node, state, slot_name)
                self.set_silent_loading_states(slot_node, state)



import json
from typing import Any, Dict

from dash import dcc, html
from dash.development.base_component import Component


class ChildContainer(html.Div):
    class ids:
        container = lambda segment: {
            "type": "DASH-ROUTER-CHILD-ROUTE-CONTAINER",
            "index": str(segment) + "-children",
        }

    class props:
        active = None

    def __init__(
        self,
        layout: Component,
        parent_segment: str,
        segment: str = None,
        *args,
        **kwargs,
    ):
        self.props.active = segment

        super().__init__(
            id=self.ids.container(parent_segment),
            children=layout,
            disable_n_clicks=True,
            *args,
            **kwargs,
        )


class SlotContainer(html.Div):
    class ids:
        container = lambda segment, slot_name: {
            "type": "DASH-ROUTER-SLOT-ROUTE-CONTAINER",
            "index": str(segment) + "-slot-" + str(slot_name),
        }

    class props:
        active = None
        is_loaded = None

    def __init__(
        self,
        layout: Component,
        parent_segment: str,
        slot_name: str,
        is_loaded: bool = True,
        *args,
        **kwargs,
    ):
        self.props.active = parent_segment
        self.props.is_loaded = is_loaded

        super().__init__(
            id=self.ids.container(parent_segment, slot_name),
            children=layout,
            disable_n_clicks=True,
            *args,
            **kwargs,
        )


class RootContainer(html.Div):
    class ids:
        container = "dash-router-root-container"
        location = "dash-router-location"
        state_store = "dash-router-loading-state-store"
        dummy = "dash-router-dummy-container"

    def __init__(self) -> None:
        super().__init__(
            [
                dcc.Location(id=self.ids.location, refresh="callback-nav"),
                html.Div(id=self.ids.container, disable_n_clicks=True),
                dcc.Store(id=self.ids.state_store, data={}),
                html.Div(id=self.ids.dummy, disable_n_clicks=True),
            ]
        )


class LacyContainer(html.Div):
    class ids:
        container = lambda index: {
            "index": index,
            "type": "dash-router-lacy-component",
        }

    def __init__(self, children: Component, node_id: str, variables: Dict[str, Any]):
        data_prop = {"data-path": json.dumps(variables)}

        super().__init__(
            children, disable_n_clicks=True, id=self.ids.container(node_id), **data_prop
        )



import importlib
import json
import os
from re import T
import traceback
import sys
from typing import Awaitable, Callable, Dict, List, Literal
from uuid import UUID, uuid4
from pathlib import Path

from dash import html
from dash._get_paths import app_strip_relative_path
from dash._utils import inputs_to_vals
from dash._validate import validate_and_group_input_args
from dash.development.base_component import Component
from flash import Flash, Input, Output, State, MATCH
from flash._pages import _parse_query_string, _infer_module_name
from quart import request
import asyncio

from .utils.constants import DEFAULT_LAYOUT_TOKEN
from .utils.helper_functions import (
    format_relative_path,
    path_to_module,
    recursive_to_plotly_json,
    format_relative_path,
    _invoke_layout,
)

from .components import ChildContainer, LacyContainer, RootContainer, SlotContainer
from .core.routing import PageNode, RouteConfig, RouteTable, RouteTree, RouterResponse
from .core.query_params import extract_function_inputs
from .core.context import RoutingContext
from .core.execution import ExecNode


class Router:
    def __init__(
        self,
        app: Flash,
        pages_folder: str = "pages",
        requests_pathname_prefix: str | None = None,
        ignore_empty_folders: bool = False,
    ) -> None:
        self.app = app
        self.static_routes = {}
        self.dynamic_routes = {}
        self.route_table = {}
        self.requests_pathname_prefix = requests_pathname_prefix
        self.ignore_empty_folders = ignore_empty_folders
        self.pages_folder = app.pages_folder if app.pages_folder else pages_folder

        if not isinstance(self.app, Flash):
            raise TypeError(f"App needs to be of Flash not: {type(self.app)}")

        self.setup_route_tree()
        self.setup_router()
        self.setup_lacy_callback()

    def setup_route_tree(self) -> None:
        """Sets up the route tree by traversing the pages folder."""
        # root_dir = ".".join(self.app.server.name.split(os.sep)[:-1])

        pages_path = Path(self.app.config.pages_folder)
        app_dir = pages_path.parent

        if not pages_path.exists():
            raise FileNotFoundError(
                f"Pages folder not found at: {pages_path}\n"
                f"Current working directory: {Path.cwd()}\n"
                f"Current working directory: {self.pages_folder}\n"
            )

        self._traverse_directory(str(app_dir), self.pages_folder, None)

    def _validate_node(self, node: PageNode):
        # Validate Slots

        # Validate children
        if node.default_child:
            pass

    def _validate_tree(self):
        for root_node in self.dynamic_routes.routes.items():
            self._validate_node(root_node)

    def _traverse_directory(
        self,
        parent_dir: str,
        segment: str,
        current_node: PageNode | None,
    ) -> None:
        """Recursively traverses the directory structure and registers routes."""
        current_dir = os.path.join(parent_dir, segment)
        if not os.path.exists(current_dir):
            return

        entries = os.listdir(current_dir)
        dir_has_page = "page.py" in entries

        if dir_has_page:
            new_node = self.load_route_module(current_dir, segment, current_node)
            if new_node is not None:

                RouteTable.add_node(new_node)
                RouteTree.add_node(new_node, current_node)

                next_node = new_node
            else:
                next_node = current_node
        else:
            next_node = current_node

        for entry in sorted(entries):
            if entry.startswith((".", "_")) or entry == "page.py":
                continue

            full_path = os.path.join(current_dir, entry)
            if os.path.isdir(full_path):
                self._traverse_directory(current_dir, entry, next_node)

    def load_route_module(
        self, current_dir: str, segment: str, parent_node: PageNode
    ) -> PageNode | None:
        """Load modules and create Page Node"""
        relative_path = os.path.relpath(current_dir, self.app.config.pages_folder)
        relative_path = format_relative_path(relative_path)
        page_module_name = path_to_module(relative_path, "page.py")
        parent_node_id = parent_node.node_id if parent_node else None

        route_config = (
            self.import_route_component(current_dir, "page.py", "config")
            or RouteConfig()
        )
        is_static = route_config.is_static or relative_path == "/"
        is_root = parent_node and parent_node.segment == "/"
        segment = relative_path if is_static else segment

        page_layout = self.import_route_component(current_dir, "page.py")
        loading_layout = self.import_route_component(current_dir, "loading.py")
        error_layout = (
            self.import_route_component(current_dir, "error.py") or self.app._on_error
        )

        endpoint = self.import_route_component(current_dir, "api.py", "endpoint")
        endpoint_inputs, end_types = extract_function_inputs(endpoint)
        layout_inputs, inp_types = extract_function_inputs(page_layout)
        inputs = set(endpoint_inputs + layout_inputs)

        node_id = relative_path # format(abs(hash(relative_path)))
        new_node = PageNode(
            _segment=segment,
            node_id=node_id,
            layout=page_layout,
            parent_id=parent_node_id,
            module=page_module_name,
            is_root=is_root,
            error=error_layout,
            loading=loading_layout,
            endpoint=endpoint,
            endpoint_inputs=inputs,
            path=relative_path,
            is_static=is_static,
            default_child=route_config.default_child,
        )

        return new_node

    def strip_relative_path(self, path: str) -> str:
        return app_strip_relative_path(self.app.config.requests_pathname_prefix, path)

    def import_route_component(
        self,
        current_dir: str,
        file_name: Literal["page.py", "error.py", "loading.py", "api.py"],
        component_name: Literal["layout", "config", "endpoint"] = "layout",
    ) -> Callable[..., Component] | Component | None:
        # page_module_name = path_to_module(current_dir, file_name)
        page_path = os.path.join(current_dir, file_name)
        page_module_name = _infer_module_name(page_path)

        try:
            spec = importlib.util.spec_from_file_location(page_module_name, page_path)
            if spec is None or spec.loader is None:
                raise ImportError(f"Cannot create spec for {page_path}")

            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            layout = getattr(module, component_name, None)

            if page_module_name not in sys.modules:
                sys.modules[page_module_name] = module            

            if file_name == "page.py" and not layout and component_name == "layout":
                raise ImportError(
                    f"Module {page_module_name} needs a layout function or component"
                )
            return layout

        except ImportError as e:
            if file_name == "page.py" and component_name == "layout":
                print(f"Error processing {page_module_name}: {e}")
                print(f"Traceback: {traceback.format_exc()}")
                raise ImportError(
                    f"Module {page_module_name} needs a layout function or component"
                )
        except Exception as e:
            print(f"Error processing {page_module_name}: {e}")
            print(f"Traceback: {traceback.format_exc()}")

        return None

    def build_execution_tree(
        self,
        current_node: PageNode,
        ctx: RoutingContext,
    ) -> ExecNode:
        """
        Recursively builds the execution tree for the matched route.
        It extracts any path variables, processes child nodes, and handles slot nodes.
        """

        if not current_node:
            return current_node

        next_segment = ctx.peek_segment()
        segment_key = current_node.create_segment_key(next_segment)
        print("next_segment", current_node.segment_value, next_segment, segment_key, flush=True)
        is_lacy = ctx.should_lazy_load(current_node, segment_key)

        if current_node.is_path_template:
            ctx.consume_path_var(current_node)
            next_segment = ctx.peek_segment()
        
        print("Current node: ", current_node.child_nodes, current_node.segment_value, current_node.path, flush=True)

        exec_node = ExecNode(
            segment=segment_key,
            node_id=current_node.node_id,
            layout=current_node.layout,
            parent_id=current_node.parent_id,
            variables=ctx.variables,
            loading=current_node.loading,
            error=current_node.error,
            is_lacy=is_lacy,
        )

        if is_lacy:
            ctx.set_node_state(current_node, "done", segment_key)
            return exec_node

        if current_node.endpoint and not DEFAULT_LAYOUT_TOKEN in segment_key:
            ctx.add_endpoint(current_node)

        if current_node.child_nodes or current_node.path_template:
            child_node = current_node.get_child_node(next_segment)

            if not child_node or not child_node.is_path_template:
                ctx.pop_segment()

            child_exec = self.build_execution_tree(
                current_node=child_node,
                ctx=ctx,
            )

            exec_node.child_node = child_exec

        if current_node.slots:
            exec_node.slots = self._process_slot_nodes(
                current_node=current_node,
                ctx=ctx,
            )

        ctx.set_node_state(current_node, "done", segment_key)
        return exec_node

    def _process_slot_nodes(
        self,
        current_node: PageNode,
        ctx: RoutingContext,
    ) -> Dict[str, ExecNode]:
        """Processes all slot nodes defined on the current node."""
        slot_exec_nodes: Dict[str, ExecNode] = {}
        for slot_name, slot_id in current_node.slots.items():
            slot_node = RouteTable.get_node(slot_id)
            segment_key = slot_node.create_segment_key(None)
            ctx.set_node_state(slot_node, "done", segment_key)

            slot_exec_node = self.build_execution_tree(
                current_node=slot_node,
                ctx=ctx,
            )

            slot_exec_nodes[slot_name] = slot_exec_node

        return slot_exec_nodes

    # ─── RESPONSE BUILDER ─────────────────────────────────────
    async def resolve_url(
        self,
        pathname: str,
        query_parameters: Dict[str, any],
        loading_state: Dict[str, any],
    ) -> RouterResponse:

        path = self.strip_relative_path(pathname)
        ctx = RoutingContext.from_request(
            pathname=path,
            query_params=query_parameters,
            loading_state_dict=loading_state,
            resolve_type="url",
        )

        static_route, path_variables = RouteTree.get_static_route(ctx)
        if static_route:
            layout = await _invoke_layout(
                static_route.layout, **query_parameters, **path_variables
            )
            return self.build_response(
                node=static_route, layout=layout, loading_state={}
            )

        active_node = RouteTree.get_active_root_node(ctx, self.ignore_empty_folders)

        if not active_node:
            return self.build_response(node=None, loading_state={})

        exec_tree = self.build_execution_tree(
            current_node=active_node,
            ctx=ctx,
        )

        if not exec_tree:
            return self.build_response(node=None, loading_state={})

        result_data = await ctx.gather_endpoints()
        final_layout = await exec_tree.execute(result_data)

        new_loading_state = {
            **ctx.get_updated_loading_state(),
            "query_params": query_parameters,
        }

        response = self.build_response(
            node=active_node, loading_state=new_loading_state, layout=final_layout
        )

        return response

    async def resolve_search(
        self,
        pathname: str,
        query_params: Dict[str, any],
        updated_query_parameters: Dict[str, any],
        loading_state: Dict[str, any],
    ) -> RouterResponse:

        path = self.strip_relative_path(pathname)
        ctx = RoutingContext.from_request(
            pathname=path,
            loading_state_dict=loading_state,
            query_params=query_params,
            resolve_type="search",
        )

        # Collect all eligible nodes (nodes whose endpoint inputs match updated query parameters)
        eligible_nodes = []
        for _, loaded_node in ctx._loading_states.items():
            node_id = loaded_node.node_id
            node = RouteTable.get_node(node_id)
            if any(
                node_input in updated_query_parameters
                for node_input in node.endpoint_inputs
            ):
                eligible_nodes.append(node)

        if not eligible_nodes:
            return

        # Find parent nodes and create a set of nodes to process
        nodes_to_process_ids = set()
        for node in eligible_nodes:
            current = node
            while current:
                # If we find a parent that needs to be processed, add it and stop traversing
                if any(
                    parent_input in updated_query_parameters
                    for parent_input in current.endpoint_inputs
                ):
                    nodes_to_process_ids.add(current.node_id)
                    break
                current = (
                    RouteTable.get_node(current.parent_id)
                    if current.parent_id
                    else None
                )

        # If no parent nodes need processing, process the eligible nodes directly
        if not nodes_to_process_ids:
            nodes_to_process_ids = {node.node_id for node in eligible_nodes}

        # Build execution trees for all selected nodes
        exec_trees = []
        nodes_to_process = []
        for node_id in nodes_to_process_ids:
            node = RouteTable.get_node(node_id)
            exec_tree = self.build_execution_tree(
                current_node=node,
                ctx=ctx,
            )
            if exec_tree:
                exec_trees.append(exec_tree)
                nodes_to_process.append(node)

        # Gather all endpoints once
        endpoint_results = await ctx.gather_endpoints()

        # Execute all trees with the same endpoint results
        layouts = []
        nodes = []
        for exec_tree, node in zip(exec_trees, nodes_to_process):
            layout = await exec_tree.execute(endpoint_results)
            if layout:
                layouts.append(layout)
                nodes.append(node)

        new_loading_state = {
            **loading_state,
            **ctx.get_updated_loading_state(),
            "query_params": query_params,
        }

        return self.build_multi_response(nodes, new_loading_state, layouts)

    @staticmethod
    async def gather_endpoints(endpoints: Dict[UUID, Callable[..., Awaitable[any]]]):
        if not endpoints:
            return {}

        keys = list(endpoints.keys())
        funcs = list(endpoints.values())
        results = await asyncio.gather(
            *[func() for func in funcs], return_exceptions=True
        )
        return dict(zip(keys, results))

    def build_response(
        self,
        node: PageNode,
        loading_state,
        layout: Component = None,
        remove_layout: bool = False,
    ):
        match node:
            case None:
                container_id = RootContainer.ids.container
                layout = html.H1("404 - Page not found")
                loading_state = {}

            case _ if node.is_root or node.is_static:
                container_id = (
                    RootContainer.ids.container
                    if not remove_layout
                    else json.dumps(ChildContainer.ids.container(node.node_id))
                )

            case _ if node.is_slot:
                container_id = json.dumps(
                    SlotContainer.ids.container(node.parent_id, node.segment)
                )

            case _:
                container_id = json.dumps(
                    ChildContainer.ids.container(
                        node.parent_id if not remove_layout else node.node_id
                    )
                )

        rendered_layout = recursive_to_plotly_json(layout)
        response = {
            container_id: {"children": rendered_layout},
            RootContainer.ids.state_store: {"data": loading_state},
        }
        return RouterResponse(multi=True, response=response)

    def build_multi_response(
        self, nodes: List[PageNode], loading_state, layouts: List[Component]
    ) -> RouterResponse:
        """Builds a response containing multiple layout updates with a single state store."""
        if not nodes or not layouts:
            return self.build_response(None, {})

        response = {}

        for node, layout in zip(nodes, layouts):
            single_response = self.build_response(node, loading_state, layout)
            response.update(single_response.response)

        return RouterResponse(multi=True, response=response)

    # ─── ASYNC & SYNC ROUTER SETUP ───────────────────────────────────────────────────
    def setup_router(self) -> None:
        @self.app.server.before_request
        async def router():
            request_data = await request.get_data()
            if not request_data:
                return

            body = json.loads(request_data)
            changed_prop = body.get("changedPropIds")

            if changed_prop:
                parts = changed_prop[0].split(".")
                changed_prop_id = parts[0]
                prop = parts[1] if len(parts) > 1 else None
            else:
                return

            if changed_prop_id != RootContainer.ids.location:
                return

            output = body["output"]
            inputs = body.get("inputs", [])
            state = body.get("state", [])
            cb_data = self.app.callback_map[output]
            inputs_state_indices = cb_data["inputs_state_indices"]
            args = inputs_to_vals(inputs + state)
            pathname_, search_, loading_state_, states_ = args
            query_parameters = _parse_query_string(search_)
            previous_qp = loading_state_.pop("query_params", {})
            _, func_kwargs = validate_and_group_input_args(args, inputs_state_indices)
            func_kwargs = dict(list(func_kwargs.items())[3:])
            varibales = {**query_parameters, **func_kwargs}

            if prop == "pathname":
                try:
                    response = await self.resolve_url(
                        pathname_, varibales, loading_state_
                    )
                    return response.model_dump()
                except Exception:
                    print(f"Traceback: {traceback.format_exc()}")
                    raise Exception("Failed to resolve the URL")

            if prop == "search":
                updated = dict(set(query_parameters.items()) - set(previous_qp.items()))
                missing_keys = previous_qp.keys() - query_parameters.keys()
                missing = {
                    key: None
                    for key in missing_keys
                    if key not in self.app.routing_callback_inputs
                }
                updates = dict(updated.items() | missing.items())
                response = await self.resolve_search(
                    pathname_, varibales, updates, loading_state_
                )
                return response.model_dump() if response else response

        @self.app.server.before_serving
        async def trigger_router():
            inputs = dict(
                pathname_=Input(RootContainer.ids.location, "pathname"),
                search_=Input(RootContainer.ids.location, "search"),
                loading_state_=State(RootContainer.ids.state_store, "data"),
            )
            inputs.update(self.app.routing_callback_inputs)

            @self.app.callback(
                Output(RootContainer.ids.dummy, "children"),
                inputs=inputs,
            )
            async def update(
                pathname_: str, search_: str, loading_state_: str, **states
            ):
                pass

    def setup_lacy_callback(self):
        inputs = dict(
            lacy_segment_id=Input(LacyContainer.ids.container(MATCH), "id"),
            variables=Input(LacyContainer.ids.container(MATCH), "data-path"),
            pathname=State(RootContainer.ids.location, "pathname"),
            search=State(RootContainer.ids.location, "search"),
            loading_state=State(RootContainer.ids.state_store, "data"),
        )

        @self.app.callback(
            Output(LacyContainer.ids.container(MATCH), "children"), inputs=inputs
        )
        async def load_lacy_component(
            lacy_segment_id, variables, pathname, search, loading_state
        ):
            print(f"Loading lacy component: {lacy_segment_id}", flush=True)
            print(f"Pathname: {pathname}", flush=True)
            print(f"Search: {search}", flush=True)
            print(f"Loading state: {loading_state}", flush=True)
            print(f"Variables: {variables}", flush=True)
            
            node_id = lacy_segment_id.get("index")
            qs = _parse_query_string(search)
            query_parameters = loading_state.pop("query_params", {})
            node_variables = json.loads(variables)
            variables = {**qs, **query_parameters, **node_variables}
            lacy_node = RouteTable.get_node(node_id)
            path = self.strip_relative_path(pathname)
            segments = path.split("/")
            node_segments = lacy_node.module.split(".")[:-1]
            current_index = node_segments.index(lacy_node.segment_value.replace("_", "-"))
            remaining_segments = list(reversed(segments[current_index:]))

            ctx = RoutingContext.from_request(
                pathname=pathname,
                query_params=variables,
                loading_state_dict=loading_state,
                resolve_type="lacy",
            )

            ctx.segments = remaining_segments

            exec_tree = self.build_execution_tree(
                current_node=lacy_node,
                ctx=ctx,
            )

            endpoint_results = await ctx.gather_endpoints()
            layout = await exec_tree.execute(endpoint_results)

            return layout
